@
@ GCat's Master System Emulator
@ Copyright (c) 2012 James Perry
@
@ This program is free software: you can redistribute it and/or modify
@ it under the terms of the GNU General Public License as published by
@ the Free Software Foundation, either version 3 of the License, or
@ (at your option) any later version.
@
@ This program is distributed in the hope that it will be useful,
@ but WITHOUT ANY WARRANTY; without even the implied warranty of
@ MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
@ GNU General Public License for more details.
@
@ You should have received a copy of the GNU General Public License
@ along with this program.  If not, see <http://www.gnu.org/licenses/>.
@

@
@ This is inlined into the end of scanline code and should generate one
@ sound sample.
@
@ r3 is the scanline number (0-311)
@ Must preserve all registers except r2 and r8
@

	strh	r5, [r0, #14]		@ free some registers
	strh	r6, [r0, #16]
	strh	r7, [r0, #18]

	ldr	r8, [r0, #88]		@ soundState pointer

	mov	r2, #0x80		@ initial sample
	
	@ do channel A first
	ldrb	r5, [r8, #0]		@ channel A volume
	rsb	r5, r5, #15		@ get right way up
	ldr	r6, [r8, #16]		@ channel A counter
	cmp	r6, #0
	addge	r2, r2, r5
	sublt	r2, r2, r5

	add	r6, r6, #15		@ update counter
	ldrh	r7, [r8, #4]		@ channel A frequency
	cmp	r7, #0
	addeq	r7, r7, #1
	cmp	r6, r7			@ counter overflowed?
	subge	r6, r6, r7, lsl #1
	str	r6, [r8, #16]		@ store counter
	

	@ do channel B
	ldrb	r5, [r8, #1]		@ channel B volume
	rsb	r5, r5, #15		@ get right way up
	ldr	r6, [r8, #20]		@ channel B counter
	cmp	r6, #0
	addge	r2, r2, r5
	sublt	r2, r2, r5

	add	r6, r6, #15		@ update counter
	ldrh	r7, [r8, #6]		@ channel B frequency
	cmp	r7, #0
	addeq	r7, r7, #1
	cmp	r6, r7			@ counter overflowed?
	subge	r6, r6, r7, lsl #1
	str	r6, [r8, #20]		@ store counter
	

	@ do channel C
	ldrb	r5, [r8, #2]		@ channel C volume
	rsb	r5, r5, #15		@ get right way up
	ldr	r6, [r8, #24]		@ channel C counter
	cmp	r6, #0
	addge	r2, r2, r5
	sublt	r2, r2, r5

	add	r6, r6, #15		@ update counter
	ldrh	r7, [r8, #8]		@ channel C frequency
	cmp	r7, #0
	addeq	r7, r7, #1
	cmp	r6, r7			@ counter overflowed?
	subge	r6, r6, r7, lsl #1
	str	r6, [r8, #24]		@ store counter

	@ noise channel
	@
	@ this has a counter like the others, but instead of toggling
	@ the output bit, it clocks a LFSR. The frequency is also set
	@ differently - it can be 16, 32, 64, or taken from channel B
	@
	@ LFSR is 16 bits. On resetting the noise channel by writing
	@ to its register, the bit 15 is set to 1, all others 0.
	@
	@ For white noise (type 1) bits 0 and 3 are XORed and fed back
	@ into bit 15.
	@
	@ For periodic noise (type 0) bit alone is fed back so the
	@ register just rotates.
	@
	@ Output is always from bit 0.
	@

	ldrb	r5, [r8, #3]		@ noise volume
	rsb	r5, r5, #15		@ get right way up
	ldrh	r6, [r8, #32]		@ get the LFSR
	tst	r6, #1
	addne	r2, r2, r5		@ add the noise output
	subeq	r2, r2, r5

	ldr	r5, [r8, #28]		@ get noise counter
	add	r5, r5, #15		@ update it
	str	r5, [r8, #28]		@ store updated noise counter
	ldrh	r7, [r8, #34]		@ get noise frequency
	cmp	r7, #0
	addeq	r7, r7, #1
	cmp	r5, r7
	blt	noise_no_update
	sub	r5, r5, r7, lsl #1	@ dec counter now
	str	r5, [r8, #28]		@ store updated noise counter

	ldrb	r5, [r8, #10]		@ read noise config
	tst	r5, #4			@ periodic noise?
	and	r7, r6, #1		@ get low bit of LFSR
	eorne	r7, r7, r6, lsr #3	@ XOR bit 3 for white noise
	and	r7, r7, #1
	mov	r6, r6, lsr #1		@ shift the LFSR
	orr	r6, r6, r7, lsl #15
	strh	r6, [r8, #32]		@ store updated LFSR
	
noise_no_update:
	
	ldr	r8, [r8, #12]		@ soundBuffer
	strb	r2, [r8, r3]		@ store sample in it
	
	ldrh	r5, [r0, #14]
	ldrh	r6, [r0, #16]
	ldrh	r7, [r0, #18]
	